#include <unistd.h>
#include <stdio.h>
#include <stdarg.h>
#include <stdlib.h>
#include <sys/types.h>
#include <sys/socket.h>
#include <sys/time.h>
#include <errno.h>
#include <string.h>
#include <netinet/in.h>
#include <arpa/inet.h>
#include <sys/uio.h>
#include <netdb.h>
#include <sys/mman.h>
#include <sys/wait.h>
#include <sys/time.h>
#include <time.h>
#include <inttypes.h>
#include <syscall.h>
#include <sys/stat.h>
#include <fcntl.h>

/*
 *
 * TODO
 *  - checksum the data some day.
 *  - use poll to wait instead of blocking recvmsg?  doesn't seem great.
 *  - measure us/call of nonblocking recvmsg
 *  - do something about receiver congestion
 *  - notice when parent tcp socket dies
 *  - should the parent be at a higher priority?
 *  - catch ctl-c
 *  - final stats summary page
 */


/* TODO: We need to read this from /proc/sys/net/rds/sol_rds */
#define SOL_RDS   272

#define RDS_CANCEL_SENT_TO      1
#define RDS_GET_MR              2
#define RDS_FREE_MR             3
#define RDS_BARRIER             4

#define RDS_CMSG_RDMA_ARGS      1

struct rds_iovec
{
  uint64_t     addr;
  uint64_t     bytes;
};

struct rds_barrier_args {
         uint32_t   daddr;
         uint32_t   flags;  /* MSG_DONTWAIT */
         uint64_t   rdma_id_addr;
         uint64_t   wait_rdma_id;
};

struct rds_get_mr_args
{
  struct rds_iovec        vec;
  uint64_t                     key_addr;
  uint64_t                     phy_addr;
};

#define RDS_FREE_MR_ARGS_INVALIDATE 1

struct rds_free_mr_args
{
  uint64_t             key;
  uint64_t             flags;
};

#define RDS_RDMA_ARGS_WRITE 1  /* read when not set */

enum {
  RDS_STRESS_RDMA_READ = 1,
  RDS_STRESS_RDMA_WRITE
};

struct rds_rdma_args
{
  struct rds_iovec        remote_vec;
  uint64_t                r_key;
  uint64_t                local_vec_addr;
  uint64_t                nr_local;
  uint64_t                flags;
  uint64_t                rdma_id_addr;

};

struct options { 
        int             performance;
        int             verbose;
	uint32_t	req_depth;
	uint32_t	req_size;
	uint32_t	iterations;
	uint32_t	ack_size;
	uint32_t	rdma_size;
	uint32_t	send_addr;
	uint32_t	receive_addr;
	uint16_t	starting_port;
	uint16_t	nr_tasks;
} __attribute__((packed));

struct counter {
	uint64_t	nr;
	uint64_t	sum;
	uint64_t	min;
	uint64_t	max;
};

enum {
	S_REQ_TX_BYTES = 0,
	S_REQ_RX_BYTES,
	S_ACK_TX_BYTES,
	S_ACK_RX_BYTES,
	S_RDMA_WRITE_BYTES,
	S_RDMA_READ_BYTES,
	S_SENDMSG_USECS,
	S_RTT_USECS,
	S__LAST
};

#define NR_STATS S__LAST

/*
 * Parents share a mapped array of these with their children.  Each child
 * gets one.  It's used to communicate between the child and the parent
 * simply.
 */
struct child_control {
	int ready;
	struct timeval start;
	struct counter cur[NR_STATS];
	struct counter last[NR_STATS];
} __attribute__((aligned (256))); /* arbitrary */

struct soak_control {
	uint64_t	per_sec;
	uint64_t	counter;
	uint64_t	last;
	struct timeval	start;
} __attribute__((aligned (256))); /* arbitrary */

/*
 * Requests tend to be larger and we try to keep a certain number of them
 * in flight at a time.  Acks are sent in response to requests and tend
 * to be smaller.
 */
#define OP_REQ 1
#define OP_ACK 2

/*
 * Every message sent with sendmsg gets a header.  This lets the receiver
 * verify that it got what was sent.
 */
struct header {
        uint64_t        rdma_addr; 
        uint64_t        rdma_phyaddr; 
        uint64_t        rdma_pattern; 
        uint64_t        rdma_key;
	uint32_t	rdma_size;
	uint32_t	req_depth;
	uint32_t	seq;
	uint32_t	from_addr;
	uint32_t	to_addr;
	uint16_t	from_port;
	uint16_t	to_port;
        uint16_t        tnr;
        uint16_t        bid;
	uint8_t		op;
	uint8_t		rdma_op;
	uint8_t		data[0];
} __attribute__((packed));

#define MIN_MSG_BYTES (4 * sizeof(struct header))
#define die(fmt...) do {		\
	fprintf(stderr, fmt);		\
	exit(1);			\
} while (0)

#define die_errno(fmt, args...) do {				\
	fprintf(stderr, fmt ", errno: %d (%s)\n", ##args , errno,\
		strerror(errno));				\
	exit(1);						\
} while (0)

#define min(a,b) (a < b ? a : b)
#define max(a,b) (a > b ? a : b)

/* global pointer to options */
static struct options *gopts = 0;

static inline rdsprintf(char *fmt, ...)
{
  va_list pvar;
  if (0 == gopts->verbose)
    return;
  va_start(pvar, fmt);
  vfprintf(stdout, fmt, pvar);
  va_end(pvar);
}

/* zero is undefined */
static inline uint64_t minz(uint64_t a, uint64_t b)
{
	if (a == 0)
		return b;
	if (b == 0)
		return a;
	return min(a, b);
}

static unsigned long long parse_ull(char *ptr, unsigned long long max)
{
	unsigned long long val;
	char *endptr;

	val = strtoull(ptr, &endptr, 0);
	if (*ptr && !*endptr && val <= max)
		return val;

	die("invalid number '%s'\n", ptr);
}

static uint32_t parse_addr(char *ptr)
{
	uint32_t addr;
        struct hostent *hent;

        hent = gethostbyname(ptr);
        if (hent && 
            hent->h_addrtype == AF_INET && hent->h_length == sizeof(addr)) {
		memcpy(&addr, hent->h_addr, sizeof(addr));
		return ntohl(addr);
	}

	die("invalid host name or dotted quad '%s'\n", ptr);
}

static void usage(void)
{
	printf(
	"Required parameters, no defaults:\n"
	" -p [port]         starting port number\n"
	" -r [addr]         receive on this host or dotted quad\n"
	" -s [addr]         send to this passive dotted quad\n"
	"\n"
	"Optional parameters, with defaults:\n"
	" -a [bytes, %u]    ack message length\n"
	" -q [bytes, 1024]  request message length\n"
	" -d [depth, 1]     request pipeline depth, nr outstanding\n"
	" -R [bytes, 0]     ack rdma data length\n"
	" -t [nr, 1]        number of child tasks\n"
	" -V [trace]        trace executions\n"
	" -D [debug]        fill and compare data\n"
	"\n"
	"Optional behavioural flags:\n"
	" -c                measure cpu use with per-cpu soak processes\n"
	"\n"
	"Example:\n"
	"  recv$ rds-stress -r recv -p 4000\n"
	"  send$ rds-stress -r send -s recv -p 4000 -q 4096 -t 2 -d 2\n"
	"\n", MIN_MSG_BYTES);

	exit(2);
}

/* This hack lets children notice when their parents die, it's not so great */
static void check_parent(char *path, pid_t pid)
{
	if (access(path, 1))
		die_errno("parent %u exited", pid);
}

/*
 * put a pattern in the message so the remote side can verify that it's
 * what was expected.  For now we put a little struct at the beginning,
 * end, and middle of the message.  We made sure that all three could fit
 * by making the min message size 128.
 */
static void fill_hdr(char *message, uint32_t bytes, struct header *hdr)
{
	uint32_t off[3] = { 0, bytes / 2, bytes - sizeof(struct header) };
	int i;

	for (i = 0; i < 3; i++)
		memcpy(message + off[i], hdr, sizeof(struct header));
}

static char *inet_ntoa_32(uint32_t val)
{
	struct in_addr addr = { .s_addr = val };
	return inet_ntoa(addr);
}

static int check_hdr(void *message, uint32_t bytes, struct header *hdr)
{
	struct header *chk;
	uint32_t off[3] = { 0, bytes / 2, bytes - sizeof(struct header) };
	int i;

#if 0
	for (i = 0; i < 3; i++) {
		chk = message + off[i];
		if (!memcmp(chk, hdr, sizeof(struct header)))
			continue;

#define bleh(var, disp)					\
		disp(hdr->var),				\
		chk->var == hdr->var ? " =" : "!=",	\
		disp(chk->var)

		/* 
		 * This is printed as one GIANT printf() so that it serializes
		 * with stdout() and we don't get things stomping on each
		 * other
		 */
		printf( "An incoming message had a header at offset %u which\n"
			"didn't contain the fields we expected:\n"
			"    member        expected eq             got\n"
			"       seq %15u %s %15u\n"
			" from_addr %15s %s %15s\n"
			" from_port %15u %s %15u\n"
			"   to_addr %15s %s %15s\n"
			"   to_port %15u %s %15u\n"
			"        op %15u %s %15u\n",
			off[i],
			bleh(seq, ntohl),
			bleh(from_addr, inet_ntoa_32),
			bleh(from_port, ntohs),
			bleh(to_addr, inet_ntoa_32),
			bleh(to_port, ntohs),
			bleh(op, (uint8_t)));
#undef bleh

		return 1;
	}
#endif
	return 0;
}

void stat_inc(struct counter *ctr, uint64_t val)
{
	ctr->nr++;
	ctr->sum += val;
	ctr->min = minz(val, ctr->min);
	ctr->max = max(val, ctr->max);
}

int64_t tv_cmp(struct timeval *a, struct timeval *b)
{
	int64_t a_usecs = ((uint64_t)a->tv_sec * 1000000ULL) + a->tv_usec;
	int64_t b_usecs = ((uint64_t)b->tv_sec * 1000000ULL) + b->tv_usec;

	return a_usecs - b_usecs;
}

/* returns a - b in usecs */
uint64_t usec_sub(struct timeval *a, struct timeval *b)
{
	return ((uint64_t)(a->tv_sec - b->tv_sec) * 1000000ULL) + 
		a->tv_usec - b->tv_usec;
}

static int bound_socket(int domain, int type, int protocol,
			struct sockaddr_in *sin)
{
	int fd;
	int opt;

	fd = socket(domain, type, protocol);
	if (fd < 0)
		die_errno("socket(%d, %d, %d) failed", domain, type, protocol);

	opt = 1;
	if (setsockopt(fd, SOL_SOCKET, SO_REUSEADDR, &opt, sizeof(opt)))
		die_errno("setsockopt(SO_REUSEADDR) failed");

	if (bind(fd, (struct sockaddr *)sin, sizeof(struct sockaddr_in)))
		die_errno("bind() failed");

	return fd;
}

static int rds_socket(struct options *opts, struct sockaddr_in *sin)
{
	char str[32] = {0, };
	int bytes;
	int fd;
	int val;
	socklen_t optlen;

	fd = open("/proc/sys/net/rds/pf_rds", O_RDONLY);
	if (fd < 0)
		die_errno("open(/proc/sys/net/rds/pf_rds) failed");

	read(fd, str, sizeof(str));

	sscanf(str, "%d", &val);
	close(fd);

	fd = bound_socket(val, SOCK_SEQPACKET, 0, sin);

	bytes = opts->nr_tasks * opts->req_depth * 
		(opts->req_size + opts->ack_size);

	if (setsockopt(fd, SOL_SOCKET, SO_SNDBUF, &bytes, sizeof(bytes)))
		die_errno("setsockopt(SNDBUF, %d) failed", bytes);
	if (setsockopt(fd, SOL_SOCKET, SO_RCVBUF, &bytes, sizeof(bytes)))
		die_errno("setsockopt(RCVBUF, %d) failed", bytes);

	optlen = sizeof(val);
	if (getsockopt(fd, SOL_SOCKET, SO_SNDBUF, &val, &optlen))
		die_errno("getsockopt(SNDBUF) failed");

#if 0	
	if (val / 2 < bytes) 
		die("getsockopt(SNDBUF) returned %d, we wanted %d * 2\n",
			val, bytes);
#endif
	optlen = sizeof(val);
	if (getsockopt(fd, SOL_SOCKET, SO_RCVBUF, &val, &optlen))
		die_errno("getsockopt(RCVBUF) failed");
#if 0
	if (val /2 < bytes) 
		die("getsockopt(RCVBUF) returned %d, we need %d * 2\n",
			val, bytes);
#endif
	return fd;
}

static uint64_t wait_for_rdma(int fd, struct sockaddr_in *saddr, uint64_t wait_rdma_id)
{
  struct rds_barrier_args args;
  uint64_t rdma_id = 0;

  args.daddr = (uint32_t)saddr->sin_addr.s_addr; 
  args.rdma_id_addr = (uint64_t)&rdma_id;
  args.wait_rdma_id = wait_rdma_id;
  args.flags = 0;

  if (setsockopt(fd, SOL_RDS, RDS_BARRIER, (char*)&args, sizeof(args)))
    die_errno("setsockopt(RDS_BARRIER) failed");

  rdsprintf("RDS wait_for_rdma rdma_id is %Lx\n",rdma_id);

  return rdma_id;

}

static uint64_t get_rdma_key(int fd, uint64_t addr, uint32_t size,  uint64_t *phyaddr)
{
  uint64_t     rkey = 0;
  struct rds_get_mr_args mr_args;

  mr_args.vec.addr = addr;
  mr_args.vec.bytes = size;
  mr_args.key_addr = (uint64_t)&rkey;
  mr_args.phy_addr = (uint64_t)phyaddr;

  if (setsockopt(fd, SOL_RDS, RDS_GET_MR, (char*)&mr_args, sizeof(mr_args)))
    die_errno("setsockopt(RDS_GET_MR) failed");

  if (*phyaddr) {
      rdsprintf("RDS get_rdma_key address is physical %Lx\n",*phyaddr);
  }

  rdsprintf("RDS get_rdma_key key is %Lx\n",rkey);

  return rkey;
}

static void free_rdma_key(int fd, uint64_t key)
{
  struct rds_free_mr_args mr_args;
  mr_args.key = key;
#if 1
  mr_args.flags = 0;
#else
  mr_args.flags = RDS_FREE_MR_ARGS_INVALIDATE;
#endif
  if (setsockopt(fd, SOL_RDS, RDS_FREE_MR, (char*)&mr_args, sizeof(mr_args)))
    die_errno("setsockopt(RDS_GET_MR) failed");
  rdsprintf("RDS free_rdma_key key is %Lx\n",key);
}


static void  rds_fill_buffer(uint64_t *addr, int size, uint64_t pattern)
{
  int d ;
  for (d = 0; d < size/sizeof(uint64_t); d++)
    addr[d] = pattern;
}


static void  rds_dump_buffer(uint64_t *addr, int size)
{
  int d ;
  for (d = 0; d < size/sizeof(uint64_t); d++)
    printf("addr %p data %llx\n",
	   &addr[d],addr[d]);
}

static void  rds_compare_buffer(uint64_t *addr, int size, uint64_t pattern)
{
  int d ;
  int failed = 0;
  for (d = 0; d < size / sizeof(uint64_t); d++)
    {
      if (addr[d] != pattern) {
	failed = 1;
	rdsprintf("compare fail pattern %Lx addr %Lx data %Lx\n",
	       (unsigned long long)pattern,(unsigned long long)&addr[d],
	       (unsigned long long)addr[d]);

#if 0
	rds_dump_buffer(addr, size);
	die("compare pass\n");
#endif
      }
    }
  if (!failed)
    rdsprintf("compare pass pattern %Lx addr %Lx\n",
	   (unsigned long long)pattern,(unsigned long long)addr);
  
}

static void run_child(pid_t parent_pid, struct child_control *ctl,
		      struct options *opts, uint16_t id)
{
  int fd;
  int fill = opts->performance ? 0 : 1;
  int compare = opts->performance ? 0 : 1;
  int verbose = opts->verbose ? 1 : 0;
  ssize_t ret;
  socklen_t socklen;
  char parent_proc[PATH_MAX];
  char buf[max(opts->req_size, opts->ack_size)];
  uint16_t i;
  uint64_t * local_buf[opts->nr_tasks][opts->req_depth];
  uint64_t * rdma_buf[opts->nr_tasks][opts->req_depth];
  uint64_t rdma_id[opts->nr_tasks][opts->req_depth];
  uint32_t depth[opts->nr_tasks];
  uint32_t buffid[opts->nr_tasks];
  uint32_t send_seq[opts->nr_tasks];
  uint32_t recv_seq[opts->nr_tasks];
  uint32_t iterations = 0;
  struct timeval send_time[opts->nr_tasks][opts->req_depth];
  struct timeval start;
  struct timeval stop;
  struct header hdr;
  struct header *in_hdr;
  struct sockaddr_in sin;

  gopts = opts;

  memset(depth, 0, sizeof(depth));
  memset(buffid, 0, sizeof(buffid));
  memset(rdma_id, 0, sizeof(rdma_id));
  memset(send_seq, 0, sizeof(send_seq));
  memset(recv_seq, 0, sizeof(recv_seq));
  memset(local_buf, 0, sizeof(local_buf));
  memset(rdma_buf, 0, sizeof(rdma_buf));
  
  sin.sin_family = AF_INET;
  sin.sin_port = htons(opts->starting_port + 1 + id);
  sin.sin_addr.s_addr = htonl(opts->receive_addr);
  
  fd = rds_socket(opts, &sin);
  
  sprintf(parent_proc, "/proc/%llu", (unsigned long long)parent_pid);
  
  ctl->ready = 1;
  
  while (ctl->start.tv_sec == 0) {
    check_parent(parent_proc, parent_pid);
    sleep(1);
  }
  
  /* sleep until we're supposed to start */ 
  gettimeofday(&start, NULL);
  if (tv_cmp(&start, &ctl->start) < 0)
    usleep(usec_sub(&ctl->start, &start));
  
  sin.sin_family = AF_INET;
  
  if (opts->rdma_size) {
    uint32_t b;
    
    /* for rdma to be page size multiple for now */
    /* this is just to simplify debugging */
    /* we need to support rdma sizes from 1 to 1meg byte */

    if (opts->rdma_size)
      opts->rdma_size = (opts->rdma_size + 4095) & ~4095;
    
    for (i = 0; i < opts->req_depth; i++)
      {
	uint64_t p = 0;
	p = (uint64_t)malloc(opts->rdma_size+4095);
	rdma_buf[id][i] = (uint64_t*)((p + 4095) & ~4095);
	rdsprintf("remote %p\n",rdma_buf[id][i]);
	memset((char*)rdma_buf[id][i],0,opts->rdma_size);
      }
  }

  while (!opts->iterations || iterations < opts->iterations) {
    
    check_parent(parent_proc, parent_pid);
    
    /* keep the pipeline full */
    for (i = 0; i < opts->nr_tasks; i++) {
      
      if (depth[i] == opts->req_depth)
	continue;
      
      sin.sin_addr.s_addr = htonl(opts->send_addr);
      sin.sin_port = htons(opts->starting_port + 1 + i);
      
      hdr.op = OP_REQ;
      hdr.seq = htonl(send_seq[i]);
      hdr.from_addr = htonl(opts->receive_addr);
      hdr.from_port = htons(opts->starting_port + 1 + id);
      hdr.to_addr = sin.sin_addr.s_addr;
      hdr.to_port = sin.sin_port;
      hdr.rdma_key = 0;
      hdr.rdma_addr = 0;
      hdr.rdma_phyaddr = 0;
      hdr.rdma_op = 0;
      hdr.rdma_pattern = 0;
      hdr.rdma_size = 0;
      hdr.req_depth = 0;
      hdr.tnr = i;

      if (opts->rdma_size && send_seq[i] > 10) {


#if 0
  {
    uint64_t rdma_key[20];
    char rdma_buff[4096*20];
    uint64_t rdma_phyaddr = 0;
    int i = 0;
    
    for (i = 0; i < 10; i++)
      rdma_key[i] = get_rdma_key(fd, &rdma_buff[i*4096], 4096, &rdma_phyaddr);
    
    for (i = 0; i < 10; i++)
      free_rdma_key(fd, rdma_key[i]);
    
  }
#endif

	hdr.bid = buffid[hdr.tnr]++ % opts->req_depth;
#if 1
	if (id & 1)
	  hdr.rdma_op = send_seq[i] & 1 ? RDS_STRESS_RDMA_WRITE : RDS_STRESS_RDMA_READ;
	else
	  hdr.rdma_op = send_seq[i] & 1 ? RDS_STRESS_RDMA_READ : RDS_STRESS_RDMA_WRITE;
#else
	hdr.rdma_op = RDS_STRESS_RDMA_READ;
#endif
	hdr.rdma_pattern = (uint64_t)((uint64_t)(send_seq[i]) << 32) | getpid();
	hdr.rdma_addr = (uint64_t)rdma_buf[hdr.tnr][hdr.bid];
	hdr.rdma_phyaddr = 0;
	hdr.rdma_size = opts->rdma_size;
	hdr.req_depth = opts->req_depth;
	hdr.rdma_key = get_rdma_key(fd, hdr.rdma_addr, hdr.rdma_size, &hdr.rdma_phyaddr);
	
	if (RDS_STRESS_RDMA_READ == hdr.rdma_op) {
	  if (fill)
	    rds_fill_buffer((uint64_t*)hdr.rdma_addr,hdr.rdma_size,hdr.rdma_pattern);
	  rdsprintf("RDS Issuing Read request for pattern %Lx local addr to rdma read %Lx\n", 
		    *((uint64_t*)(hdr.rdma_addr)),
		    hdr.rdma_addr);
	} else {
	  
	  if (fill)
	    rds_fill_buffer((uint64_t*)hdr.rdma_addr,hdr.rdma_size,0);
	  rdsprintf("RDS Issuing Write request for pattern %Lx local addr to rdma write %Lx\n", 
		    hdr.rdma_pattern,
		    hdr.rdma_addr);

	}
      } /* rdma */      
      fill_hdr(buf, opts->req_size, &hdr);
      
      gettimeofday(&start, NULL);
      send_time[i][send_seq[i] % opts->req_depth] = start;
      
      ret = sendto(fd, buf, opts->req_size, 0,
		   (struct sockaddr *)&sin, sizeof(sin));
      
      gettimeofday(&stop, NULL);
      if (ret != opts->req_size)
	die_errno("sendto() returned %zd", ret);
      stat_inc(&ctl->cur[S_REQ_TX_BYTES], ret);
      stat_inc(&ctl->cur[S_SENDMSG_USECS], 
	       usec_sub(&stop, &start));
      
      depth[i]++;
      send_seq[i]++;
      
    } /* for each task */
      
    /* 
     * we've filled the pipeline, so block receiving until
     * an ack clears space for us to send again or we recv
     * a message.
     */
    socklen = sizeof(struct sockaddr_in);
    ret = recvfrom(fd, buf, max(opts->req_size, opts->ack_size), 0, 
		   (struct sockaddr *)&sin, &socklen);
    if (ret < sizeof(struct header))
      die_errno("recvfrom() returned %zd", ret);
    if (socklen < sizeof(struct sockaddr_in))
      die("socklen = %d < sizeof(sin) (%zu)\n",
	  socklen, sizeof(struct sockaddr_in));
    
    /* make sure the incoming message's size matches is op */
    in_hdr = (void *)buf;
    switch(in_hdr->op) {
    case OP_REQ:
      stat_inc(&ctl->cur[S_REQ_RX_BYTES], ret);
      if (ret != opts->req_size)
	die("req size %zd, not %u\n", ret,
	    opts->req_size);
      break;
    case OP_ACK:
      stat_inc(&ctl->cur[S_ACK_RX_BYTES], ret);
      if (ret != opts->ack_size)
	die("ack size %zd, not %u\n", ret,
	    opts->ack_size);
      break;
    default:
      die("unknown op %u\n", in_hdr->op);
    }
    
    /* check the incoming sequence number */
    i = ntohs(sin.sin_port) - opts->starting_port - 1;
    
    /*
     * Verify that the incoming header indicates that this 
     * is the next in-order message to us.  We can't predict
     * op.
     */
    hdr.op = in_hdr->op;
    hdr.seq = htonl(recv_seq[in_hdr->tnr]);
    hdr.from_addr = sin.sin_addr.s_addr;
    hdr.from_port = sin.sin_port;
    hdr.to_addr = htonl(opts->receive_addr);
    hdr.to_port = htons(opts->starting_port + 1 + id);
    
    if (hdr.op == OP_ACK) {
      gettimeofday(&stop, NULL);
      stat_inc(&ctl->cur[S_RTT_USECS], 
	       usec_sub(&stop,
			&send_time[in_hdr->tnr][recv_seq[in_hdr->tnr] % opts->req_depth]));
    }
    
    recv_seq[in_hdr->tnr]++;
    
    if (check_hdr(buf, ret, &hdr))
      die("header from %s:%u to id %u bogus\n",
	  inet_ntoa(sin.sin_addr), htons(sin.sin_port),
	  id);
    
    if (hdr.op == OP_ACK) {
      
      depth[in_hdr->tnr]--;
      
      if (in_hdr->rdma_key) {
	
	rdsprintf("RDS rcvd rdma %s completion for request key %Lx len %u local addr %Lx phyaddr %Lx\n",
		  RDS_STRESS_RDMA_WRITE == in_hdr->rdma_op ? "write" : "read",
		  in_hdr->rdma_key,in_hdr->rdma_size,in_hdr->rdma_addr,in_hdr->rdma_phyaddr);
	
	free_rdma_key(fd,in_hdr->rdma_key);
	
	/* if acking an rdma write request - then remote node wrote local host buffer */
	/* (data in) so count this as rdma data coming in (rdma_read) - else remote node read */
	/* local host buffer so count this as rdma write (data out) */

	if (RDS_STRESS_RDMA_WRITE == in_hdr->rdma_op) {
	  /* remote node wrote local buffer check pattern */
	  /* sent via immediate data in rdma buffer */
	  
	  stat_inc(&ctl->cur[S_RDMA_READ_BYTES],  in_hdr->rdma_size);

	  if (compare)
	    rds_compare_buffer((uint64_t*)in_hdr->rdma_addr, in_hdr->rdma_size, 
			       in_hdr->rdma_pattern);
	}
	else
	  stat_inc(&ctl->cur[S_RDMA_WRITE_BYTES],  in_hdr->rdma_size);

      iterations++;
      } /* rdma */

      if (!opts->rdma_size)
	iterations++;

      continue;
    } /* ack */
    
    /* send an ack in response to the req we just got */
    /* if rdma enabled, issue the rdma with immediate data */
    hdr.op = OP_ACK;
    hdr.seq = htonl(send_seq[in_hdr->tnr]);
    hdr.from_addr = htonl(opts->receive_addr);
    hdr.from_port = htons(opts->starting_port + 1 + id);
    hdr.to_addr = sin.sin_addr.s_addr;
    hdr.to_port = sin.sin_port;
    hdr.rdma_size = in_hdr->rdma_size;
    hdr.req_depth = in_hdr->req_depth;
    hdr.rdma_key = in_hdr->rdma_key; 
    hdr.rdma_phyaddr = in_hdr->rdma_phyaddr; /* remote's address to rdma to / from */
    hdr.rdma_addr = in_hdr->rdma_addr; /* remote's address to rdma to / from */
    hdr.tnr = in_hdr->tnr;
    hdr.bid = in_hdr->bid;
    
    fill_hdr(buf, opts->ack_size, &hdr);
    
    stat_inc(&ctl->cur[S_ACK_TX_BYTES], ret);
    
    if (0 == in_hdr->rdma_key) {
      
      ret = sendto(fd, buf, opts->ack_size, 0,
		   (struct sockaddr *)&sin, sizeof(sin));
      if (ret != opts->ack_size)
	die_errno("sendto() returned %zd", ret);
      
    } else {
      
      int rdma_id_slot = 0;
      uint64_t done_rdma_id;
      uint64_t barrier_rdma_id = 0;
#define MSG_MAXIOVLEN 2
      struct rds_iovec   localva[MSG_MAXIOVLEN];
      struct iovec   immdva[MSG_MAXIOVLEN];
      struct cmsghdr *cmsg;
      char ctlbuf[CMSG_SPACE(sizeof(struct rds_rdma_args))];
      struct rds_rdma_args *rdmap;
      struct msghdr msg;
      
      rdsprintf("RDS recv'd request to issue rdma %s len %u rva %Lx rpa %Lx pattern %Lx\n",
		in_hdr->rdma_op == RDS_STRESS_RDMA_WRITE ? "write to" : "read from",
		in_hdr->rdma_size, in_hdr->rdma_addr, in_hdr->rdma_phyaddr, in_hdr->rdma_pattern);

      /* allocate local buffers if not already done */
      if (0 == local_buf[in_hdr->tnr][0]) {
	int ii;
	for (ii = 0; ii < in_hdr->req_depth; ii++)
	  {
	    uint64_t p = 0;
	    p = (uint64_t)malloc(in_hdr->rdma_size+4095);
	    local_buf[in_hdr->tnr][ii] = (uint64_t*)((p + 4095) & ~4095);
	    rdsprintf("local %p\n",local_buf[in_hdr->tnr][ii]);
	    memset((char*)local_buf[in_hdr->tnr][ii],0,in_hdr->rdma_size);
	  }
      }
      
      /* find free rdma_id slot */
      rdma_id_slot = recv_seq[in_hdr->tnr] % in_hdr->req_depth;
      
      if (rdma_id[in_hdr->tnr][rdma_id_slot]) {
	rdsprintf("waiting for rdma to complete %Lx slot %u\n",
		  rdma_id[in_hdr->tnr][rdma_id_slot],rdma_id_slot);
	done_rdma_id = wait_for_rdma(fd,&sin,rdma_id[in_hdr->tnr][rdma_id_slot]);
	if (done_rdma_id < rdma_id[in_hdr->tnr][rdma_id_slot]) {
	  die("WAIT FAILED %Lx to complete, barrier is %Lx slot %u\n",
	      rdma_id[in_hdr->tnr][rdma_id_slot],done_rdma_id,rdma_id_slot);
	}
	else {
	  int d; 
	  /* free all slots completed */
	  for (d = 0; d < in_hdr->req_depth; d++) {
	    if (rdma_id[in_hdr->tnr][d] <= done_rdma_id)
	      rdma_id[in_hdr->tnr][d] = 0;
	  }
	}
      }
      if (rdma_id[in_hdr->tnr][rdma_id_slot])
	die("Ran out of rdma id slots -> slot %u seq %u rdma_id[slot] %Lx\n", 
	    rdma_id_slot,recv_seq[in_hdr->tnr], rdma_id[in_hdr->tnr][rdma_id_slot]);
      
      rdsprintf("using rdma_id_slot %u\n",rdma_id_slot);

      /* cmsg setup for rdma */
      msg.msg_control = ctlbuf;
      msg.msg_controllen = sizeof(ctlbuf);
      cmsg = CMSG_FIRSTHDR(&msg);
      cmsg->cmsg_level = SOL_RDS;
      cmsg->cmsg_type = RDS_CMSG_RDMA_ARGS;
      cmsg->cmsg_len = CMSG_LEN(sizeof(struct rds_rdma_args));
      msg.msg_controllen = cmsg->cmsg_len;
      
      /* local buffer to rdma in to / out of */ 
      localva[0].addr = local_buf[in_hdr->tnr][in_hdr->bid]; 
      localva[0].bytes = in_hdr->rdma_size;
      
      /* rdma args */
      rdmap = (struct rds_rdma_args *)CMSG_DATA(cmsg);
      rdmap->local_vec_addr = localva;
      rdmap->nr_local = 1;
      rdmap->rdma_id_addr = &rdma_id[in_hdr->tnr][rdma_id_slot];
      if (in_hdr->rdma_phyaddr)
	rdmap->remote_vec.addr =  in_hdr->rdma_phyaddr;
      else
	rdmap->remote_vec.addr =  0; /* zero based fmrs ....in_hdr->rdma_addr */
      rdmap->remote_vec.bytes = in_hdr->rdma_size;
      rdmap->r_key = in_hdr->rdma_key;
      
      /* read or write */
      if (RDS_STRESS_RDMA_WRITE == in_hdr->rdma_op) {
	if (fill)
	  rds_fill_buffer((uint64_t*)localva[0].addr,in_hdr->rdma_size,in_hdr->rdma_pattern);
	rdmap->flags = RDS_RDMA_ARGS_WRITE;

	/* need separate rdma stats cells */
	stat_inc(&ctl->cur[S_RDMA_WRITE_BYTES], in_hdr->rdma_size);
      
      }
      else
	{
	  stat_inc(&ctl->cur[S_RDMA_READ_BYTES], in_hdr->rdma_size);
	  rdmap->flags = 0;
	}

      /* immediate data (ack message) to go */
      immdva[in_hdr->tnr].iov_base = buf;
      immdva[in_hdr->tnr].iov_len = opts->ack_size;
      
      /* sendmsg immediate data header */
      msg.msg_iovlen = 1;
      msg.msg_iov =  immdva;
      msg.msg_name  = (struct sockaddr *)&sin;
      msg.msg_namelen  = sizeof(sin);
      
      rdsprintf("RDS issuing rdma for key %Lx rva %Lx len %Lu lva %lx\n",
		hdr.rdma_key,rdmap->remote_vec.addr,rdmap->remote_vec.bytes,localva[0].addr);

      /*issue rdma with immediate data following */

      ret = sendmsg(fd, &msg, 0);
      
      if (ret != opts->ack_size)
	die_errno("sendto() returned %zd", ret);
      
      rdsprintf("RDS rdma operation Id is %Lx\n",rdma_id[in_hdr->tnr][rdma_id_slot]);
      
      gettimeofday(&stop, NULL);
      
      
    }
  }
}
  
static struct child_control *start_children(struct options *opts)
    {
	struct child_control *ctl;
	pid_t parent = getpid();
	pid_t pid;
	size_t len;
	uint32_t i;

	len = opts->nr_tasks * sizeof(*ctl);
	ctl = mmap(NULL, len, PROT_READ|PROT_WRITE, MAP_ANONYMOUS|MAP_SHARED,
		   0, 0);
	if (ctl == MAP_FAILED)
		die("mmap of %u child control structs failed", opts->nr_tasks);

	memset(ctl, 0, len);

	for (i = 0; i < opts->nr_tasks; i++) {
		pid = fork();
		if (pid == -1)
			die_errno("forking child nr %u failed", i);
		if (pid == 0) {
			run_child(parent, ctl + i, opts, i);
			exit(0);
		}
	}

	for (i = 0; i < opts->nr_tasks; i++) {
		if (ctl[i].ready)
			continue;
		pid = waitpid(-1, NULL, WNOHANG);
		if (pid)
			die("child %u (pid %u) exited\n", i, pid);
		sleep(1);
		i--; /* try this child again */
	}

	return ctl;
}

static double avg(struct counter *ctr)
{
	if (ctr->nr)
		return (double)ctr->sum / (double)ctr->nr;
	else
		return 0.0;
}

static double throughput(struct counter *disp)
{
	return disp[S_REQ_TX_BYTES].sum + disp[S_REQ_RX_BYTES].sum +
	  disp[S_ACK_TX_BYTES].sum + disp[S_ACK_RX_BYTES].sum;
}

static double throughput_rdma(struct counter *disp)
{
        return 	 disp[S_RDMA_WRITE_BYTES].sum + disp[S_RDMA_READ_BYTES].sum;
}

void stat_snapshot(struct counter *disp, struct child_control *ctl,
		   uint16_t nr_tasks)
{
	struct counter tmp[NR_STATS];
	uint16_t i;
	uint16_t s;

	memset(disp, 0, sizeof(tmp));

	for (i = 0; i < nr_tasks; i++) {
		memcpy(tmp, ctl[i].cur, sizeof(tmp));

		for (s = 0; s < NR_STATS; s++) {
			disp[s].nr += tmp[s].nr - ctl[i].last[s].nr;
			disp[s].sum += tmp[s].sum - ctl[i].last[s].sum;
			disp[s].min = minz(tmp[s].min, ctl[i].last[s].min);
			disp[s].max = max(tmp[s].max, ctl[i].last[s].max);
		}

		memcpy(ctl[i].last, tmp, sizeof(tmp));
	}
}

void stat_total(struct counter *disp, struct child_control *ctl,
		uint16_t nr_tasks)
{
	uint16_t i;
	uint16_t s;

	memset(disp, 0, sizeof(struct counter) * NR_STATS);

	for (i = 0; i < nr_tasks; i++) {
		for (s = 0; s < NR_STATS; s++) {
			disp[s].nr += ctl[i].cur[s].nr;
			disp[s].sum += ctl[i].cur[s].sum;
			disp[s].min = minz(disp[s].min, ctl[i].cur[s].min);
			disp[s].max = max(disp[s].max, ctl[i].cur[s].max);
		}
	}
}

static double cpu_use(struct soak_control *soak_arr)
{
	struct soak_control *soak;
	uint64_t capacity = 0;
	uint64_t soaked = 0;
	uint64_t this;

	if (soak_arr == NULL)
		return -1.0;

	for (soak = soak_arr; soak && soak->per_sec; soak++) {
		capacity += soak->per_sec;
		this = soak->counter;
		soaked += min(soak->per_sec, this - soak->last);
		soak->last = this;
	}

	return (double)(capacity - soaked) * 100 / (double)capacity;
}

static void release_children_and_wait(struct options *opts,
				      struct child_control *ctl,
				      struct soak_control *soak_arr)
{
	struct counter disp[NR_STATS];
	struct timeval start;
	int status;
	pid_t pid;
	uint16_t i;
	uint16_t nr_running;

	gettimeofday(&start, NULL);
	start.tv_sec += 2;
	for (i = 0; i < opts->nr_tasks; i++)
		ctl[i].start = start;

	nr_running = opts->nr_tasks;

	printf("%4s %6s %10s %10s %7s %8s %5s\n",
		"tsks", "tx/s", "tx+rx K/s", "rw+rr K/s", "tx us/c", "rtt us", "cpu %");

	while (nr_running) {
		/* XXX big bug, need to mark some ctl elements dead */
		stat_snapshot(disp, ctl, nr_running);

		sleep(1);

		printf("%4u %6"PRIu64" %10.2f %10.2f %7.2f %8.2f %5.2f\n",
			nr_running,
			disp[S_REQ_TX_BYTES].nr,
			throughput(disp) / 1000.0,
			throughput_rdma(disp) / 1000.0,
			avg(&disp[S_SENDMSG_USECS]),
			avg(&disp[S_RTT_USECS]),
			cpu_use(soak_arr));

		/* see if any children have finished or died */
		pid = waitpid(-1, &status, WNOHANG);
		if (pid < 0)
			die("waitpid returned %u", pid);
		if (pid == 0)
			continue;
		if (WIFEXITED(status) && WEXITSTATUS(status) == 0) {
			nr_running--;
			continue;
		}

		die("child pid %u exited with status %d\n",
		    pid, WEXITSTATUS(status));
	}

	stat_total(disp, ctl, nr_running);
	printf("request sent: %"PRIu64"\n", disp[S_REQ_TX_BYTES].nr);
}	

static int active_parent(struct options *opts, struct soak_control *soak_arr)
{
	struct child_control *ctl;
	struct sockaddr_in sin;
	int fd;
	ssize_t ret;
	uint8_t ok;

	ctl = start_children(opts);

	sin.sin_family = AF_INET;
	sin.sin_port = htons(opts->starting_port);
	sin.sin_addr.s_addr = htonl(opts->receive_addr);

	fd = bound_socket(PF_INET, SOCK_STREAM, IPPROTO_TCP, &sin);

	sin.sin_family = AF_INET;
	sin.sin_port = htons(opts->starting_port);
	sin.sin_addr.s_addr = htonl(opts->send_addr);

	printf("connecting to %s:%u\n", inet_ntoa(sin.sin_addr),
		ntohs(sin.sin_port));

	if (connect(fd, (struct sockaddr *)&sin, sizeof(sin)))
		die_errno("connect() failed");

	printf("connected to %s:%u\n", inet_ntoa(sin.sin_addr),
		ntohs(sin.sin_port));

	ret = write(fd, opts, sizeof(struct options));
	if (ret != sizeof(struct options))
		die_errno("option write returned %zd", ret);

	ret = read(fd, &ok, sizeof(ok));
	if (ret != sizeof(ok))
		die_errno("ok read returned %zd", ret);

	printf("negotiated options, tasks will start in 2 seconds\n");

	release_children_and_wait(opts, ctl, soak_arr);

	return 0;
}

static int passive_parent(uint32_t addr, uint16_t port,
			  struct soak_control *soak_arr)
{
	struct options remote, *opts = &remote;
	struct child_control *ctl;
	struct sockaddr_in sin;
	socklen_t socklen;
	int fd;
	ssize_t ret;
	uint8_t ok;

	sin.sin_family = AF_INET;
	sin.sin_port = htons(port);
	sin.sin_addr.s_addr = htonl(addr);

	fd = bound_socket(PF_INET, SOCK_STREAM, IPPROTO_TCP, &sin);

	if (listen(fd, 255))
		die_errno("listen() failed");

	fd = accept(fd, (struct sockaddr *)&sin, &socklen);
	if (fd < 0)
		die_errno("accept() failed");

	printf("accepted connection from %s:%u\n", inet_ntoa(sin.sin_addr),
		ntohs(sin.sin_port));

	ret = read(fd, opts, sizeof(struct options));
	if (ret != sizeof(struct options))
		die_errno("remote option read returned %zd\n", ret);

	/* 
	 * The sender gave us their send and receive addresses, we need
	 * to swap them.
	 */
	opts->send_addr = opts->receive_addr;
	opts->receive_addr = addr;

	ctl = start_children(opts);

	ret = write(fd, &ok, sizeof(ok));
	if (ret != sizeof(ok))
		die_errno("remote option read returned %zd\n", ret);

	printf("negotiated options, tasks will start in 2 seconds\n");

	release_children_and_wait(opts, ctl, soak_arr);

	return 0;
}

/*
 * The soaker *constantly* spins calling getpid().  It tries to execute a
 * second's worth of calls before checking that it's parent is still alive.  It
 * uses gettimeofday() to figure out the per-second rate of the series it just
 * executed.  It always tries to work from the highest rate it ever saw.  
 */
static void run_soaker(pid_t parent_pid, struct soak_control *soak)
{
	char parent_proc[PATH_MAX];
	uint64_t i;
	uint64_t per_sec;
	struct timeval start;
	struct timeval stop;
	uint64_t usecs;

	soak->per_sec = 1000;

	sprintf(parent_proc, "/proc/%llu", (unsigned long long)parent_pid);

	while (1) {
		gettimeofday(&start, NULL);
		for (i = 0; i < soak->per_sec; i++) {
			syscall(SYS_getpid);
			soak->counter++;
		}
		gettimeofday(&stop, NULL);

		usecs = usec_sub(&stop, &start);
		per_sec = (double)soak->per_sec * 1000000.0 / (double)usecs;

		if (per_sec > soak->per_sec)
			soak->per_sec = per_sec;

		check_parent(parent_proc, parent_pid);
	}
}

struct soak_control *start_soakers(void)
{
	struct soak_control *soak_arr;
	pid_t parent = getpid();
	pid_t pid;
	size_t len;
	long nr_soak = sysconf(_SC_NPROCESSORS_ONLN);
	long i;

	/* an extra terminating entry which will be all 0s */
	len = (nr_soak + 1) * sizeof(struct soak_control);
	soak_arr = mmap(NULL, len, PROT_READ|PROT_WRITE,
			MAP_ANONYMOUS|MAP_SHARED, 0, 0);
	if (soak_arr == MAP_FAILED)
		die("mmap of %ld soak control structs failed", nr_soak);

	memset(soak_arr, 0, len);

	printf("started %ld cycle soaking processes\n", nr_soak);

	for (i = 0; i < nr_soak; i++) {
		pid = fork();
		if (pid == -1)
			die_errno("forking soaker nr %lu failed", i);
		if (pid == 0) {
			run_soaker(parent, soak_arr + i);
			exit(0);
		}
	}

	return soak_arr;
}

void check_size(uint32_t size, uint32_t unspec, uint32_t max, char *desc,
		char *opt)
{
	if (size == ~0)
		die("specify %s with %s\n", desc, opt);
	if (size < max)
		die("%s must be at least %zu bytes\n", desc, max);
}

int main(int argc, char **argv)
{
	struct options opts;
	struct soak_control *soak_arr = NULL;

	memset(&opts, 0xff, sizeof(opts));

	if (argc == 1)
		usage();

	opts.ack_size = MIN_MSG_BYTES;
	opts.req_size = 1024;
	opts.rdma_size = 0;
	opts.performance = 1;
	opts.verbose = 0;

        while(1) {
		int c;

                c = getopt(argc, argv, "+a:cd:Dhi:p:q:r:R:s:t:V");
                if (c == -1)
                        break;

                switch(c) {
                        case 'a':
				opts.ack_size = parse_ull(optarg, (uint32_t)~0);
                                break;
                        case 'c':
				soak_arr = start_soakers();
                                break;
                        case 'd':
				opts.req_depth = parse_ull(optarg,(uint32_t)~0);
                                break;
                        case 'D':
				opts.performance = 0;
                                break;
                        case 'i':
				opts.iterations = parse_ull(optarg, (uint32_t)~0);
                                break;
                        case 'p':
				opts.starting_port = parse_ull(optarg,
							       (uint16_t)~0);
                                break;
                        case 'q':
				opts.req_size = parse_ull(optarg, (uint32_t)~0);
                                break;
                        case 'r':
				opts.receive_addr = parse_addr(optarg);
                                break;
                        case 'R':
				opts.rdma_size = parse_ull(optarg, (uint32_t)~0);
                                break;
                        case 's':
				opts.send_addr = parse_addr(optarg);
                                break;
                        case 't':
				opts.nr_tasks = parse_ull(optarg,
							  (uint16_t)~0);
                                break;
                        case 'V':
				opts.verbose = 1;
                                break;
                        case 'h':
                        case '?':
                        default:
				usage();
				break;
                }
        }

	if (opts.starting_port == (uint16_t)~0)
		die("specify starting port with -p\n");
	if (opts.receive_addr == ~0)
		die("specify receiving addr with -r\n");

	/* the passive parent will read options off the wire */
	if (opts.send_addr == ~0)
		return passive_parent(opts.receive_addr, opts.starting_port,
				      soak_arr);

	/* the active parent verifies and sends its options */
	check_size(opts.ack_size, ~0, MIN_MSG_BYTES, "ack size", "-a");
	check_size(opts.req_size, ~0, MIN_MSG_BYTES, "req size", "-q");

	/* defaults */
	if (opts.req_depth == ~0)
		opts.req_depth = 1;
	if (opts.nr_tasks == (uint16_t)~0)
		opts.nr_tasks = 1;

	return active_parent(&opts, soak_arr);
}

/*
 * This are completely stupid.  options.c should be removed.
 */
void print_usage(int durr) { }
void print_version() { }
